import torch

from cca_zoo.deep._discriminative._dcca_ey import DCCA_EY


class DCCA_SVD(DCCA_EY):
    """

    References
    ----------
    Chapman, James, Ana Lawry Aguila, and Lennie Wells. "A Generalized EigenGame with Extensions to Multiview Representation Learning." arXiv preprint arXiv:2211.11323 (2022).
    """

    def __init__(self, latent_dimensions: int, encoders=None, r: float = 0, **kwargs):
        super().__init__(
            latent_dimensions=latent_dimensions, encoders=encoders, r=r, **kwargs
        )
        self.r = r
        # check if the number of views is equal to 2
        if len(self.encoders) != 2:
            raise ValueError(
                f"Expected 2 views, got {len(self.encoders)} views instead."
            )

    def loss(self, batch, **kwargs):
        # views here is a list of 'paired' views (i.e. [view1, view2])
        z = self(batch["views"])  # get the latent representations
        C = torch.cov(torch.hstack(z).T)
        latent_dims = z[0].shape[1]

        Cxy = C[:latent_dims, latent_dims:]
        Cxx = C[:latent_dims, :latent_dims]

        independent_views = batch.get("independent_views", None)
        if independent_views is None:
            Cyy = C[latent_dims:, latent_dims:]
        else:
            independent_z = self(independent_views)
            Cyy = torch.cov(torch.hstack(independent_z).T)[latent_dims:, latent_dims:]

        rewards = torch.trace(2 * Cxy)
        penalties = torch.trace(Cxx @ Cyy)
        return {
            "objective": -rewards + penalties,  # return the negative objective value
            "rewards": rewards,  # return the total rewards
            "penalties": penalties,  # return the penalties matrix
        }



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
from sklearn.utils import check_matplotlib_support

from cca_zoo.utils.check_values import check_seaborn_support


class CovarianceHeatmapDisplay:
    """Covariance Heatmap Display

    Heatmap of the covariances between the latent variables of the views.

    Parameters
    ----------
    train_covariances : np.ndarray
        The train covariances between views.
    test_covariances : np.ndarray
        The test covariances between views.

    Attributes
    ----------
    figure_ : matplotlib.pyplot.figure
        The figure of the plot.

    Examples
    --------
    >>> from cca_zoo.visualisation import CovarianceHeatmapDisplay
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> from cca_zoo.linear import MCCA
    >>>
    >>> # Generate Sample Data
    >>> # --------------------
    >>> X = np.random.rand(100, 10)
    >>> Y = np.random.rand(100, 10)
    >>>
    >>> # Splitting the data into training and testing sets
    >>> X_train, X_test = X[:50], X[50:]
    >>> Y_train, Y_test = Y[:50], Y[50:]
    >>>
    >>> views = [X_train, Y_train]
    >>> test_views = [X_test, Y_test]
    >>>
    >>> # Train an MCCA Model
    >>> # -------------------
    >>> mcca = MCCA(latent_dimensions=2)
    >>> mcca.fit(views)
    >>>
    >>> # %%
    >>> # Plotting the Covariance Heatmap
    >>> # -------------------------------
    >>> CovarianceHeatmapDisplay.from_estimator(mcca, views, test_views=test_views).plot()
    >>> plt.show()

    """

    def __init__(self, train_covariances, test_covariances):
        self.train_covariances = train_covariances
        self.test_covariances = test_covariances

    def _validate_plot_params(self):
        check_seaborn_support("CorrelationHeatmapDisplay")

    @classmethod
    def from_estimator(cls, model, train_views, test_views=None):
        train_scores = model.transform(train_views)
        if test_views is not None:
            test_scores = model.transform(test_views)
        else:
            test_scores = None
        train_covariances = np.cov(train_scores[0].T, train_scores[1].T)
        if test_scores is not None:
            test_covariances = np.cov(test_scores[0].T, test_scores[1].T)
        else:
            test_covariances = None
        return cls.from_covariances(train_covariances, test_covariances)

    @classmethod
    def from_covariances(cls, train_covariances, test_covariances=None):
        return cls(train_covariances, test_covariances)

    def plot(self):
        self._validate_plot_params()
        fig, axs = plt.subplots(1, 2, figsize=(10, 5))
        sns.heatmap(
            self.train_covariances,
            annot=True,
            cmap="coolwarm",
            ax=axs[0],
        )
        if self.test_covariances is not None:
            sns.heatmap(
                self.test_covariances,
                annot=True,
                cmap="coolwarm",
                ax=axs[1],
            )
        axs[0].set_title("Train Covariances")
        axs[1].set_title("Test Covariances")
        # plt.tight_layout()
        self.figure_ = fig
        return self



import torch

from cca_zoo.deep import objectives
from cca_zoo.deep._base import BaseDeep
from cca_zoo.linear._mcca import MCCA


class DCCA(BaseDeep):
    """
    A class used to fit a DCCA model.

    References
    ----------
    Andrew, Galen, et al. "Deep canonical correlation analysis." International conference on machine learning. PMLR, 2013.

    """

    def __init__(
        self,
        latent_dimensions: int,
        objective=objectives.MCCA,
        encoders=None,
        r: float = 0,
        eps: float = 1e-5,
        **kwargs,
    ):
        super().__init__(latent_dimensions=latent_dimensions, **kwargs)
        # Check if encoders are provided and have the same length as the number of views
        if encoders is None:
            raise ValueError(
                "Encoders must be a list of torch.nn.Module with length equal to the number of views."
            )
        self.encoders = torch.nn.ModuleList(encoders)
        self.objective = objective(r=r, eps=eps)

    def forward(self, views, **kwargs):
        if not hasattr(self, "n_views_"):
            self.n_views_ = len(views)
        # Use list comprehension to encode each view
        z = [encoder(view) for encoder, view in zip(self.encoders, views)]
        return z

    def loss(self, batch, **kwargs):
        z = self(batch["views"])
        return {"objective": self.objective.loss(z)}

    def pairwise_correlations(self, loader: torch.utils.data.DataLoader):
        # Call the parent class method
        return super().pairwise_correlations(loader)

    def correlation_captured(self, z):
        # Remove mean from each view
        z = [zi - zi.mean(0) for zi in z]
        return MCCA(latent_dimensions=self.latent_dimensions).fit(z).score(z).sum()

    def score(self, loader: torch.utils.data.DataLoader, **kwargs):
        z = self.transform(loader)
        corr = self.correlation_captured(z)
        return corr



from typing import Iterable, Union

import numpy as np

from cca_zoo.linear._mcca import MCCA
from cca_zoo.utils.check_values import _process_parameter


class GCCA(MCCA):
    r"""
    A class used to fit GCCA model. This model extends CCA to more than two views by optimizing the sum of correlations with a shared auxiliary vector.

    The objective function of GCCA is:

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{ \sum_iw_i^TX_i^TT  \}\\

        \text{subject to:}

        T^TT=1

    where :math:`T` is the auxiliary vector.

    Examples
    --------
    >>> from cca_zoo.linear import GCCA
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> X3 = rng.random((10,5))
    >>> model = GCCA()
    >>> model.fit((X1,X2,X3)).score((X1,X2,X3))

    References
    ----------
    Tenenhaus, Arthur, and Michel Tenenhaus. "Regularized generalized canonical correlation analysis." Psychometrika 76.2 (2011): 257.
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        c: Union[Iterable[float], float] = None,
        view_weights: Iterable[float] = None,
        eps: float = 1e-6,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            accept_sparse=["csc", "csr"],
            random_state=random_state,
            c=c,
            eps=eps,
            pca=False,
        )
        self.view_weights = view_weights

    def fit(self, views: Iterable[np.ndarray], y=None, K=None, **kwargs):
        return super().fit(views, y=y, K=K, **kwargs)

    def _check_params(self):
        self.c = _process_parameter("c", self.c, 0, self.n_views_)

    def _C(self, views, K=None):
        if K is None:
            # just use identity when all rows are observed in all views.
            K = np.ones((len(views), views[0].shape[0]))
        Q = []
        self.view_weights = _process_parameter(
            "view_weights", self.view_weights, 1, self.n_views_
        )
        for i, (view, view_weight) in enumerate(zip(views, self.view_weights)):
            view_cov = (1 - self.c[i]) * np.cov(view, rowvar=False) + self.c[
                i
            ] * np.eye(view.shape[1])
            smallest_eig = min(0, np.linalg.eigvalsh(view_cov).min()) - self.eps
            view_cov = view_cov - smallest_eig * np.eye(view_cov.shape[0])
            Q.append(view_weight * view @ np.linalg.inv(view_cov) @ view.T)
        Q = np.sum(Q, axis=0)
        Q = (
            np.diag(np.sqrt(np.sum(K, axis=0)))
            @ Q
            @ np.diag(np.sqrt(np.sum(K, axis=0)))
        )
        return Q

    def _D(self, views, **kwargs):
        return None

    def _weights(self, eigvals, eigvecs, views, **kwargs):
        self.weights = [
            np.linalg.pinv(view) @ eigvecs[:, : self.latent_dimensions]
            for view in views
        ]

    def _more_tags(self):
        return {"multiview": True}



# from ._altmaxvar import AltMaxVar
from ._elastic import SCCA_IPLS, ElasticCCA
from ._pls_als import PLS_ALS

from ._scca_parkhomenko import SCCA_Parkhomenko
from ._scca_pmd import SCCA_PMD

from ._scca_span import SCCA_Span

__all__ = [
    "ElasticCCA",
    "SCCA_IPLS",
    "PLS_ALS",
    "SCCA_PMD",
    "SCCA_Parkhomenko",
    "SCCA_Span",
    # "AltMaxVar",
]



from typing import Union

import numpy as np

from cca_zoo.linear._iterative._base import BaseIterative
from cca_zoo.linear._iterative._deflation import DeflationMixin


class PLS_ALS(DeflationMixin, BaseIterative):
    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        tol=1e-3,
        accept_sparse=None,
        epochs=100,
        initialization: Union[str, callable] = "random",
        early_stopping=False,
        verbose=True,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            tol=tol,
            accept_sparse=accept_sparse,
            epochs=epochs,
            initialization=initialization,
            early_stopping=early_stopping,
            verbose=verbose,
        )

    def _update_weights(self, views: np.ndarray, i: int):
        # Update the weights for the current view using PLS
        # Get the scores of all views
        scores = np.stack(self.transform(views))
        # Create a mask that is True for elements not equal to i along dim i
        mask = np.arange(scores.shape[0]) != i
        # Apply the mask to scores and sum along dim i
        target = np.sum(scores[mask], axis=0)
        # Compute the new weights by computing the covariance between the view and the target
        new_weights = np.cov(np.hstack((views[i], target)).T)[:-1, -1]
        # Normalize the new weights
        new_weights /= np.linalg.norm(new_weights)
        # Return the new weights
        return new_weights[:, np.newaxis]



import torch

from cca_zoo.linear._gradient._ey import PLS_EY
from cca_zoo.linear._pls import PLSMixin


class PLSStochasticPower(PLS_EY, PLSMixin):
    def _more_tags(self):
        return {"multiview": True, "stochastic": True}

    def training_step(self, batch, batch_idx):
        if batch is None:
            batch = dict(("views", self.data))
        for weight in self.torch_weights:
            weight.data = self._orth(weight)
        scores = self(batch["views"])
        # find the pairwise covariance between the scores
        cov = torch.cov(torch.hstack(scores).T)
        loss = torch.trace(cov[: scores[0].shape[1], scores[0].shape[1] :])
        self.log("train_loss", loss, prog_bar=True)
        return loss

    def on_train_end(self) -> None:
        for weight in self.torch_weights:
            weight.data = self._orth(weight)

    @staticmethod
    def _orth(U):
        Qu, Ru = torch.linalg.qr(U)
        Su = torch.sign(torch.sign(torch.diag(Ru)) + 0.5)
        return Qu @ torch.diag(Su)



import torch

from ._dcca import DCCA
from ..objectives import inv_sqrtm


class DCCA_NOI(DCCA):
    """
    A class used to fit a DCCA model by non-linear orthogonal iterations


    References
    ----------
    Wang, Weiran, et al. "Stochastic optimization for deep CCA via nonlinear orthogonal iterations." 2015 53rd Annual Allerton Conference on Communication, Control, and Computing (Allerton). IEEE, 2015.

    """

    def __init__(
        self,
        latent_dimensions: int,
        encoders=None,
        r: float = 0,
        rho: float = 0.2,
        eps: float = 1e-9,
        **kwargs,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            encoders=encoders,
            r=r,
            eps=eps,
            **kwargs,
        )
        if rho < 0 or rho > 1:
            raise ValueError(f"rho should be between 0 and 1. rho={rho}")
        self.eps = eps
        self.rho = rho
        self.mse = torch.nn.MSELoss(reduction="sum")

    def on_fit_start(self) -> None:
        self.covs = [
            torch.eye(self.latent_dimensions, requires_grad=False, device=self.device)
        ] * 2

    def on_train_batch_start(self, batch, batch_idx):
        z = self(batch["views"])
        self._update_covariances(z)

    def loss(self, batch, **kwargs):
        z = self(batch["views"])
        covariance_inv = [inv_sqrtm(cov, self.eps) for cov in self.covs]
        preds = [z_ @ covariance_inv[i] for i, z_ in enumerate(z)]
        loss = self.mse(z[0], preds[1].detach()) + self.mse(z[1], preds[0].detach())
        return {"objective": loss}

    def _update_covariances(self, z):
        batch_covs = [torch.cov(z_.T) for z_ in z]
        self.covs = [
            self.rho * self.covs[i].detach() + (1 - self.rho) * batch_cov
            for i, batch_cov in enumerate(batch_covs)
        ]



import itertools
from abc import abstractmethod
from typing import Iterable, Union, Any

import numpy as np
from cca_zoo.linear._pls import MPLS
from tqdm import tqdm

from cca_zoo._base import BaseModel
from cca_zoo.linear._dummy import DummyCCA, DummyPLS
from cca_zoo.linear._mcca import MCCA


class BaseIterative(BaseModel):
    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        tol=1e-3,
        accept_sparse=None,
        epochs=100,
        initialization: Union[str, callable] = "random",
        early_stopping=False,
        verbose=True,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            accept_sparse=accept_sparse,
        )
        self.tol = tol
        self.epochs = epochs
        # validate the initialization method
        if initialization not in ["random", "uniform", "unregularized", "pls"]:
            raise ValueError(
                "Initialization method must be one of ['random', 'uniform', 'unregularized', 'pls']"
            )
        else:
            self.initialization = initialization
        # validate the callbacks
        self.verbose = verbose
        # validate the convergence checking
        self.early_stopping = early_stopping

    def fit(self, views: Iterable[np.ndarray], y=None, **kwargs):
        self.weights = self._fit(views)
        return self

    def _fit(self, views: Iterable[np.ndarray]):
        views = self._validate_data(views)
        self._initialize(views)
        self._check_params()
        # Solve using alternating optimisation across the views until convergence
        # Initialize the loss and the previous weights
        loss = np.inf
        prev_weights = self.weights.copy()
        # Loop over the epochs
        for epoch in tqdm(
            range(self.epochs),
            desc="Epochs",
            position=0,
            leave=True,
            disable=not self.verbose,
        ):
            # Loop over the views
            for i in range(len(views)):
                # Update the weights for the current view by solving a linear system
                self.weights[i] = self._update_weights(views, i)
            # Check if the loss has decreased enough
            curr_loss = self._objective(views)
            if self.early_stopping:
                weight_diff = np.sum(
                    [
                        np.linalg.norm(w - pw)
                        for w, pw in zip(self.weights, prev_weights)
                    ]
                ) / len(self.weights)
                if weight_diff < self.tol:
                    print(f"Early stopping at epoch {epoch}")
                    break
        # Return the final weights
        return self.weights

    @abstractmethod
    def _update_weights(self, view: np.ndarray, i: int):
        """Update the CCA weights for a given view.

        Parameters
        ----------
        view : np.ndarray
            The input view to update the CCA weights for
        i : int
            The index of the view

        Returns
        -------
        np.ndarray
            The updated CCA weights for the view
        """
        pass

    def _objective(self, views: Iterable[np.ndarray]):
        # Compute the objective function value for a given set of views using SCCA
        # Get the scores of all views
        transformed_views = self.transform(views)
        all_covs = []
        # Sum all the pairwise covariances except self covariance
        for x, y in itertools.product(transformed_views, repeat=2):
            all_covs.append(
                np.diag(
                    np.corrcoef(x.T, y.T)[
                        : self.latent_dimensions, self.latent_dimensions :
                    ]
                )
            )
        # the sum of covariances
        return np.sum(all_covs)

    def _initialize(self, views: Iterable[np.ndarray]):
        """Initialize the CCA weights using the initialization method or function.

        Parameters
        ----------
        views : Iterable[np.ndarray]
            The input views to initialize the CCA weights from
        """
        pls = self._get_tags().get("pls", False)
        initializer = _default_initializer(
            self.initialization, self.random_state, self.latent_dimensions, pls
        )
        # Fit the initializer on the input views and get the weights as numpy arrays
        self.weights = initializer.fit(views).weights
        self.weights = [weights.astype(np.float32) for weights in self.weights]

    def _more_tags(self):
        # Indicate that this class is for multiview data
        return {"iterative": True}


def _default_initializer(
    initialization: str, random_state: Any, latent_dims: int, pls: bool
) -> Union[DummyCCA, DummyPLS, MPLS, MCCA]:
    initializer_map = {
        "random": (DummyPLS if pls else DummyCCA)(
            latent_dims, random_state=random_state, uniform=False
        ),
        "uniform": (DummyPLS if pls else DummyCCA)(
            latent_dims, random_state=random_state, uniform=True
        ),
        "unregularized": (MPLS if pls else MCCA)(
            latent_dims, random_state=random_state
        ),
        "pls": MPLS(
            latent_dims, random_state=random_state
        ),  # Assuming PLS initialization can be used for both pls=True/False
    }

    initializer = initializer_map.get(initialization)

    if not initializer:
        raise ValueError(
            f"Initialization {initialization} not supported. Pass a generator implementing this method"
        )

    return initializer



from typing import Iterable, Union

import numpy as np
from scipy.linalg import block_diag
from sklearn.metrics import pairwise_kernels
from sklearn.utils.validation import check_is_fitted

from cca_zoo.linear._gcca import GCCA
from cca_zoo.linear._mcca import MCCA
from cca_zoo.linear._tcca import TCCA
from cca_zoo.utils import _process_parameter


class KernelMixin:
    def _check_params(self):
        self.kernel = _process_parameter("kernel", self.kernel, "linear", self.n_views_)
        self.gamma = _process_parameter("gamma", self.gamma, None, self.n_views_)
        self.coef0 = _process_parameter("coef0", self.coef0, 1, self.n_views_)
        self.degree = _process_parameter("degree", self.degree, 1, self.n_views_)
        self.c = _process_parameter("c", self.c, 0, self.n_views_)
        self.kernel_params = _process_parameter(
            "kernel_params", self.kernel_params, {}, self.n_views_
        )

    def _process_data(self, views, K=None):
        self.train_views = views
        kernels = [
            pairwise_kernels(
                view,
                metric=self.kernel[i],
                gamma=self.gamma[i],
                degree=self.degree[i],
                coef0=self.coef0[i],
                filter_params=True,
                **self.kernel_params[i]
            )
            for i, view in enumerate(self.train_views)
        ]
        return kernels

    def transform(self, views: Iterable[np.ndarray], **kwargs):
        check_is_fitted(self, attributes=["alphas"])
        Ktest = [
            pairwise_kernels(
                self.train_views[i],
                Y=view,
                metric=self.kernel[i],
                gamma=self.gamma[i],
                degree=self.degree[i],
                coef0=self.coef0[i],
                filter_params=True,
                **self.kernel_params[i]
            )
            for i, view in enumerate(views)
        ]
        transformed_views = [
            kernel.T @ self.alphas[i] for i, kernel in enumerate(Ktest)
        ]
        return transformed_views

    @property
    def alphas(self):
        check_is_fitted(self, attributes=["weights"])
        return self.weights

    def _more_tags(self):
        # Indicate that this class is for multiview data
        return {"kernel": True}


class KCCA(KernelMixin, MCCA):
    r"""
    A class used to fit KCCA model. This model extends MCCA to nonlinear relationships by using kernel functions on each view.

    The objective function of KCCA is:

    .. math::

        \alpha_{opt}=\underset{\alpha}{\mathrm{argmax}}\{\sum_i\sum_{j\neq i} \alpha_i^TK_i^TK_j\alpha_j  \}\\

        \text{subject to:}

        c_i\alpha_i^TK_i\alpha_i + (1-c_i)\alpha_i^TK_i^TK_i\alpha_i=1

    where :math:`K_i` are the kernel matrices for each view and :math:`c_i` are the regularization parameters for each view.

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to use, by default 1
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random seed for reproducibility, by default None
    c : Union[Iterable[float], float], optional
        Regularization parameter or list of parameters for each view, by default None. If None, it will be set to zero for each view.
    eps : float, optional
        Small value to add to the diagonal of the kernel matrices, by default 1e-3
    kernel: Iterable[Union[float, callable]], optional
        Kernel function or list of functions for each view, by default None. If None, it will use a linear kernel for each view.
    gamma: Iterable[float], optional
        Gamma parameter or list of parameters for the RBF kernel for each view, by default None. Ignored if kernel is not RBF.
    degree: Iterable[float], optional
        Degree parameter or list of parameters for the polynomial kernel for each view, by default None. Ignored if kernel is not polynomial.
    coef0: Iterable[float], optional
        Coef0 parameter or list of parameters for the polynomial or sigmoid kernel for each view, by default None. Ignored if kernel is not polynomial or sigmoid.
    kernel_params: Iterable[dict], optional
        Additional parameters or list of parameters for the kernel function for each view, by default None.

    Examples
    --------
    >>> from cca_zoo.linear import KCCA
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> X3 = rng.random((10,5))
    >>> model = KCCA()
    >>> model.fit((X1,X2,X3)).score((X1,X2,X3))
    array([0.96893666])
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        c: Union[Iterable[float], float] = None,
        eps=1e-3,
        kernel: Iterable[Union[str, float, callable]] = None,
        gamma: Iterable[float] = None,
        degree: Iterable[float] = None,
        coef0: Iterable[float] = None,
        kernel_params: Iterable[dict] = None,
    ):
        # Call the parent class constructor
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            c=c,
            eps=eps,
            pca=False,
        )
        # Store the kernel parameters
        self.kernel_params = kernel_params
        self.gamma = gamma
        self.coef0 = coef0
        self.kernel = kernel
        self.degree = degree

    def _D(self, views, **kwargs):
        D = block_diag(
            *[
                (1 - self.c[i]) * np.cov(view, rowvar=False) + self.c[i] * view
                for i, view in enumerate(views)
            ]
        )
        D_smallest_eig = min(0, np.linalg.eigvalsh(D).min()) - self.eps
        D = D - D_smallest_eig * np.eye(D.shape[0])
        return D / len(views)


class KGCCA(KernelMixin, GCCA):
    r"""
    A class used to fit KGCCA model. This model extends GCCA to nonlinear relationships by using kernel functions on each view.

    The objective function of KGCCA is:

    .. math::

    \alpha_{opt}=\underset{\alpha}{\mathrm{argmax}}\{ \sum_i\alpha_i^TK_i^TT  \}\\

    \text{subject to:}

    T^TT=1

    where :math:`K_i` are the kernel matrices for each view and :math:`T` is the auxiliary vector.

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to use, by default 1
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random seed for reproducibility, by default None
    c : Union[Iterable[float], float], optional
        Regularization parameter or list of parameters for each view, by default None. If None, it will be set to zero for each view.
    kernel: Iterable[Union[float, callable]], optional
        Kernel function or list of functions for each view, by default None. If None, it will use a linear kernel for each view.
    gamma: Iterable[float], optional
        Gamma parameter or list of parameters for the RBF kernel for each view, by default None. Ignored if kernel is not RBF.
    degree: Iterable[float], optional
        Degree parameter or list of parameters for the polynomial kernel for each view, by default None. Ignored if kernel is not polynomial.
    coef0: Iterable[float], optional
        Coef0 parameter or list of parameters for the polynomial or sigmoid kernel for each view, by default None. Ignored if kernel is not polynomial or sigmoid.
    kernel_params: Iterable[dict], optional
        Additional parameters or list of parameters for the kernel function for each view, by default None.
    view_weights : Iterable[float], optional
        Weights for each view in the objective function, by default None. If None, it will use equal weights for each view.

    References
    ----------
    Tenenhaus, Arthur, Cathy Philippe, and Vincent Frouin. "Kernel generalized canonical correlation analysis." Computational Statistics & Data Analysis 90 (2015): 114-131.

    Examples
    --------
    >>> from cca_zoo.linear import KGCCA
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> X3 = rng.random((10,5))
    >>> model = KGCCA()
    >>> model.fit((X1,X2,X3)).score((X1,X2,X3))
    array([0.97019284])
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        c: Union[Iterable[float], float] = None,
        kernel: Iterable[Union[float, callable]] = None,
        gamma: Iterable[float] = None,
        degree: Iterable[float] = None,
        coef0: Iterable[float] = None,
        kernel_params: Iterable[dict] = None,
        view_weights: Iterable[float] = None,
        eps: float = 1e-6,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            c=c,
            view_weights=view_weights,
            eps=eps,
        )
        self.kernel_params = kernel_params
        self.gamma = gamma
        self.coef0 = coef0
        self.kernel = kernel
        self.degree = degree

    def _weights(self, eigvals, eigvecs, views, **kwargs):
        kernels = [
            pairwise_kernels(
                view,
                metric=self.kernel[i],
                gamma=self.gamma[i],
                degree=self.degree[i],
                coef0=self.coef0[i],
                filter_params=True,
                **self.kernel_params[i]
            )
            for i, view in enumerate(self.train_views)
        ]
        self.weights = [
            np.linalg.pinv(kernel) @ eigvecs[:, : self.latent_dimensions]
            for kernel in kernels
        ]


class KTCCA(KernelMixin, TCCA):
    r"""
    A class used to fit KTCCA model. This model extends TCCA to nonlinear relationships by using kernel functions on each view.

    The objective function of KTCCA is:

    .. math::

        \alpha_{opt}=\underset{\alpha}{\mathrm{argmax}}\{ \alpha_1^TK_1^T\otimes \alpha_2^TK_2^T\otimes \cdots \otimes \alpha_m^TK_m^T\alpha  \}\\

        \text{subject to:}

        c_i\alpha_i^TK_i\alpha_i + (1-c_i)\alpha_i^TK_i^TK_i\alpha_i=1

    where :math:`K_i` are the kernel matrices for each view and :math:`c_i` are the regularization parameters for each view.

    References
    ----------
    Kim, Tae-Kyun, Shu-Fai Wong, and Roberto Cipolla. "Tensor canonical correlation analysis for action classification." 2007 IEEE Conference on Computer Vision and Pattern Recognition. IEEE, 2007

    Examples
    --------
    >>> from cca_zoo.linear import KTCCA
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> X3 = rng.random((10,5))
    >>> model = KTCCA()
    >>> model.fit((X1,X2,X3)).score((X1,X2,X3))
    array([1.69896269])
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        eps=1e-3,
        c: Union[Iterable[float], float] = None,
        kernel: Iterable[Union[float, callable]] = None,
        gamma: Iterable[float] = None,
        degree: Iterable[float] = None,
        coef0: Iterable[float] = None,
        kernel_params: Iterable[dict] = None,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            eps=eps,
            c=c,
        )
        self.kernel_params = kernel_params
        self.gamma = gamma
        self.coef0 = coef0
        self.kernel = kernel
        self.degree = degree

    def _setup_tensor(self, views: Iterable[np.ndarray], **kwargs):
        self.train_views = views
        kernels = [
            pairwise_kernels(
                view,
                metric=self.kernel[i],
                gamma=self.gamma[i],
                degree=self.degree[i],
                coef0=self.coef0[i],
                filter_params=True,
                **self.kernel_params[i]
            )
            for i, view in enumerate(self.train_views)
        ]
        return super()._setup_tensor(kernels)



# from typing import Iterable, Union
#
# import numpy as np
#
# from cca_zoo.utils import _process_parameter
#
# from ._base import BaseIterative, BaseLoop
# from ._deflation import DeflationMixin
#
#
# class SCCA_ADMM(BaseIterative, DeflationMixin):
#     r"""
#     Fits a sparse CCA model by alternating ADMM for two or more views.
#
#     .. math::
#
#         w_{opt}=\underset{w}{\mathrm{argmax}}\{\sum_i\sum_{j\neq i} \|X_iw_i-X_jw_j\|^2 + \|w_i\|_1\}\\
#
#         \text{subject to:}
#
#         w_i^TX_i^TX_iw_i=1
#
#     Parameters
#     ----------
#     latent_dimensions : int, default=1
#         Number of latent dimensions to use in the model.
#     copy_data : bool, default=True
#         Whether to copy the data or overwrite it.
#     random_state : int, default=None
#         Random seed for initialisation.
#     deflation : str, default="cca"
#         Deflation method to use. Options are "cca" and "pls".
#     tau : float or list of floats, default=None
#         Regularisation parameter. If a single float is given, the same value is used for all views.
#         If a list of floats is given, the values are used for each view.
#     mu : float or list of floats, default=None
#         Regularisation parameter. If a single float is given, the same value is used for all views.
#         If a list of floats is given, the values are used for each view.
#     lam : float or list of floats, default=None
#         Regularisation parameter. If a single float is given, the same value is used for all views.
#         If a list of floats is given, the values are used for each view.
#     eta : float or list of floats, default=None
#         Regularisation parameter. If a single float is given, the same value is used for all views.
#         If a list of floats is given, the values are used for each view.
#     tol : float, default=1e-9
#         Tolerance for convergence.
#
#     References
#     ----------
#     Suo, Xiaotong, et al. "Sparse canonical correlation analysis." arXiv preprint arXiv:1705.10865 (2017).
#
#     Examples
#     --------
#     >>> from cca_zoo.linear import SCCA_ADMM
#     >>> import numpy as np
#     >>> rng=np.random.RandomState(0)
#     >>> X1 = rng.random((10,5))
#     >>> X2 = rng.random((10,5))
#     >>> model = SCCA_ADMM(random_state=0,tau=[1e-1,1e-1])
#     >>> model.fit((X1,X2)).score((X1,X2))
#     array([0.84348183])
#     """
#
#     def __init__(
#         self,
#         latent_dimensions: int = 1,
#         copy_data=True,
#         random_state=None,
#         tau: Union[Iterable[float], float] = None,
#         mu: Union[Iterable[float], float] = None,
#         lam: Union[Iterable[float], float] = None,
#         eta: Union[Iterable[float], float] = None,
#         initialization: Union[str, callable] = "pls",
#         tol: float = 1e-3,
#     ):
#         self.tau = tau
#         self.mu = mu
#         self.lam = lam
#         self.eta = eta
#         super().__init__(
#             latent_dimensions=latent_dimensions,
#             copy_data=copy_data,
#             initialization=initialization,
#             tol=tol,
#             random_state=random_state,
#         )
#
#     def _check_params(self):
#         self.tau = _process_parameter("tau", self.tau, 0, self.n_views_)
#         self.lam = _process_parameter("lam", self.lam, 1, self.n_views_)
#         self.eta = _process_parameter("eta", self.eta, 0, self.n_views_)
#
#     def _get_pl_module(self, weights=None, k=None):
#         return SCCA_ADMM_PL(
#             weights=weights,
#             k=k,
#             tau=self.tau,
#             lam=self.lam,
#             eta=self.eta,
#             n_samples_=self.n_samples_,
#             n_views_=self.n_views_,
#         )
#
#
# class SCCA_ADMM_PL(BaseLoop):
#     def __init__(
#         self,
#         weights,
#         k=None,
#         n_samples_=None,
#         n_views_=None,
#         tau=None,
#         eta=None,
#         lam=None,
#         mu=None,
#     ):
#         super().__init__(weights=weights, k=k)
#         self.eta = [np.ones(n_samples_) * eta for eta in eta]
#         self.z = [np.ones(n_samples_)] * n_views_
#         self.mu = mu
#
#     def training_step(self, batch, batch_idx):
#         views = batch["views"]
#         scores = np.stack(self(views))
#         for view_index, view in enumerate(views):
#             targets = np.ma.array(scores, mask=False)
#             targets.mask[view_index] = True
#             gradient = views[view_index].T @ targets.sum(axis=0).filled()
#             mu = self.mu[view_index]
#             lam = self.lam[view_index]
#             N = views[view_index].shape[0]
#             unnorm_z = []
#             norm_eta = []
#             norm_weights = []
#             norm_proj = []
#             for _ in range(self.max_iter):
#                 # We multiply 'tau' by N in order to make regularisation match across the different sparse cca methods
#                 self.weights[view_index] = self._prox_mu_f(
#                     self.weights[view_index]
#                     - mu
#                     / lam
#                     * views[view_index].T
#                     @ (
#                         views[view_index] @ self.weights[view_index]
#                         - self.z[view_index]
#                         + self.eta[view_index]
#                     ),
#                     mu,
#                     gradient,
#                     N * self.tau[view_index],
#                 )
#                 unnorm_z.append(
#                     np.linalg.norm(
#                         views[view_index] @ self.weights[view_index]
#                         + self.eta[view_index]
#                     )
#                 )
#                 self.z[view_index] = self._prox_lam_g(
#                     views[view_index] @ self.weights[view_index] + self.eta[view_index]
#                 )
#                 self.eta[view_index] = (
#                     self.eta[view_index]
#                     + views[view_index] @ self.weights[view_index]
#                     - self.z[view_index]
#                 )
#                 norm_eta.append(np.linalg.norm(self.eta[view_index]))
#                 norm_proj.append(
#                     np.linalg.norm(views[view_index] @ self.weights[view_index])
#                 )
#                 norm_weights.append(np.linalg.norm(self.weights[view_index], 1))
#
#     def _prox_mu_f(self, x, mu, c, tau):
#         u_update = x.copy()
#         mask_1 = x + (mu * c) > mu * tau
#         # if mask_1.sum()>0:
#         u_update[mask_1] = x[mask_1] + mu * (c[mask_1] - tau)
#         mask_2 = x + (mu * c) < -mu * tau
#         # if mask_2.sum() > 0:
#         u_update[mask_2] = x[mask_2] + mu * (c[mask_2] + tau)
#         mask_3 = ~(mask_1 | mask_2)
#         u_update[mask_3] = 0
#         return u_update
#
#     def _prox_lam_g(self, x):
#         norm = np.linalg.norm(x)
#         if norm < 1:
#             return x
#         else:
#             return x / max(1, norm)



from typing import Iterable
import numpy as np
import torch
from cca_zoo.data.deep import NumpyDataset
from cca_zoo.linear._gradient._base import BaseGradientModel, FullBatchDataset
from cca_zoo.linear._pls import PLSMixin


class CCA_EY(BaseGradientModel):
    def _more_tags(self):
        return {"multiview": True, "stochastic": True}

    def training_step(self, batch, batch_idx):
        if self.batch_size is None:
            batch = self.batch
        loss = self.loss(batch["views"], batch.get("independent_views", None))
        # Logging the loss components with "train/" prefix
        for k, v in loss.items():
            self.log(
                f"train/{k}",
                v,
                prog_bar=True,
                on_epoch=True,
                batch_size=batch["views"][0].shape[0],
            )
        return loss

    def validation_step(self, batch, batch_idx):
        loss = self.loss(batch["views"], batch.get("independent_views", None))
        # Logging the loss components
        for k, v in loss.items():
            self.log(
                f"val/{k}",
                v,
                prog_bar=True,
                on_epoch=True,
                batch_size=batch["views"][0].shape[0],
            )
        return loss

    def get_AB(self, z):
        latent_dims = z[0].shape[1]
        A = torch.zeros(
            latent_dims, latent_dims, device=z[0].device
        )  # initialize the cross-covariance matrix
        B = torch.zeros(
            latent_dims, latent_dims, device=z[0].device
        )  # initialize the auto-covariance matrix
        for i, zi in enumerate(z):
            for j, zj in enumerate(z):
                if i == j:
                    B += self._cross_covariance(zi, zj, latent_dims)
                else:
                    A += self._cross_covariance(zi, zj, latent_dims)
        return A / len(z), B / len(z)

    @staticmethod
    def _cross_covariance(zi, zj, latent_dims) -> torch.Tensor:
        return torch.cov(torch.hstack((zi, zj)).T)[latent_dims:, :latent_dims]

    def loss(self, views, independent_views=None, **kwargs):
        # Encoding the views with the forward method
        z = self(views)
        # Getting A and B matrices from z
        A, B = self.get_AB(z)
        if independent_views is None:
            independent_B = B
        else:
            # Encoding another set of views with the forward method
            independent_z = self(independent_views)
            # Getting A' and B' matrices from independent_z
            independent_A, independent_B = self.get_AB(independent_z)
        # Computing rewards and penalties using A and B'
        rewards = torch.trace(2 * A)
        penalties = torch.trace(B @ independent_B)

        return {
            "loss": -rewards + penalties,
            "rewards": rewards,
            "penalties": penalties,
        }

    def get_dataset(self, views: Iterable[np.ndarray], validation_views=None):
        dataset = (
            DoubleNumpyDataset(views) if self.batch_size else FullBatchDataset(views)
        )
        if validation_views is not None:
            val_dataset = (
                DoubleNumpyDataset(validation_views)
                if self.batch_size
                else FullBatchDataset(validation_views)
            )
        else:
            val_dataset = None
        return dataset, val_dataset


class DoubleNumpyDataset(NumpyDataset):
    random_state = np.random.RandomState(0)

    def __getitem__(self, index):
        views = [view[index] for view in self.views]
        independent_index = self.random_state.randint(0, len(self))
        independent_views = [view[independent_index] for view in self.views]
        return {"views": views, "independent_views": independent_views}


class PLS_EY(CCA_EY, PLSMixin):
    def get_AB(self, z):
        latent_dims = z[0].shape[1]
        A = torch.zeros(
            latent_dims, latent_dims, device=z[0].device
        )  # initialize the cross-covariance matrix
        B = torch.zeros(
            latent_dims, latent_dims, device=z[0].device
        )  # initialize the auto-covariance matrix
        n = z[0].shape[0]
        for i, zi in enumerate(z):
            for j, zj in enumerate(z):
                if i == j:
                    B += self.torch_weights[i].T @ self.torch_weights[i] / n
                else:
                    A += self._cross_covariance(zi, zj, latent_dims)
        return A / len(z), B / len(z)



from typing import Iterable

import numpy as np

from cca_zoo._base import BaseModel


class DummyCCA(BaseModel):
    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        accept_sparse=None,
        uniform=False,
    ):
        if accept_sparse is None:
            accept_sparse = ["csc", "csr"]
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            accept_sparse=accept_sparse,
            random_state=random_state,
        )
        self.uniform = uniform

    def fit(self, views: Iterable[np.ndarray], y=None, **kwargs):
        self._validate_data(views)
        if self.uniform:
            self.weights = [
                np.ones((view.shape[1], self.latent_dimensions)) for view in views
            ]
        else:
            self.weights = [
                self.random_state.normal(
                    0, 1, size=(view.shape[1], self.latent_dimensions)
                )
                for view in views
            ]
        self.normalize_weights(views)
        return self

    def normalize_weights(self, views):
        self.weights = [
            weight
            / np.sqrt(np.diag(np.atleast_1d(np.cov(view @ weight, rowvar=False))))
            for view, weight in zip(views, self.weights)
        ]
        scores = self.score(views)
        for i, score in enumerate(scores):
            if score < 0:
                # flip the sign of the first weights
                self.weights[0][:, i] = -self.weights[0][:, i]


class DummyPLS(DummyCCA):
    def normalize_weights(self, views):
        self.weights = [
            weight / np.linalg.norm(weight, axis=0)
            for view, weight in zip(views, self.weights)
        ]



from ._dcca import DCCA
from .. import objectives


class DTCCA(DCCA):
    """
    A class used to fit a DTCCA model.

    Is just a thin wrapper round DCCA with the DTCCA objective

    References
    ----------
    Wong, Hok Shing, et al. "Deep Tensor CCA for Multi-view Learning." IEEE Transactions on Big Data (2021).

    """

    def __init__(
        self,
        latent_dimensions: int,
        encoders=None,
        r: float = 0,
        eps: float = 1e-5,
        **kwargs
    ):
        # Call the parent class constructor with the DTCCA objective function
        super().__init__(
            latent_dimensions=latent_dimensions,
            objective=objectives.TCCA,
            encoders=encoders,
            r=r,
            eps=eps,
            **kwargs
        )



from typing import Union, Iterable

import numpy as np

from cca_zoo.linear._iterative._base import BaseIterative
from cca_zoo.linear._iterative._deflation import DeflationMixin
from cca_zoo.utils import _process_parameter


class SCCA_Parkhomenko(DeflationMixin, BaseIterative):
    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        tol=1e-3,
        accept_sparse=None,
        epochs=100,
        initialization: Union[str, callable] = "pls",
        early_stopping=False,
        verbose=True,
        tau=None,  # regularization parameter for Parkhomenko
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            tol=tol,
            accept_sparse=accept_sparse,
            epochs=epochs,
            initialization=initialization,
            early_stopping=early_stopping,
            verbose=verbose,
        )
        self.tau = tau

    def _check_params(self):
        self.tau = _process_parameter("tau", self.tau, 0.0001, self.n_views_)
        if any(tau <= 0 for tau in self.tau):
            raise (
                "All regularisation parameters should be above 0. " f"tau=[{self.tau}]"
            )

    def _update_weights(self, views: Iterable[np.ndarray], i: int):
        # Update the weights for the current view using Parkhomenko
        # Get the scores of all views
        scores = np.stack(self.transform(views))
        # Create a mask that is True for elements not equal to i along dim i
        mask = np.arange(scores.shape[0]) != i
        # Apply the mask to scores and sum along dim i
        target = np.sum(scores[mask], axis=0)
        # Compute the new weights by multiplying the view with the target and dividing by the norm of the new weights
        new_weights = views[i].T @ target / np.linalg.norm(views[i].T @ target)
        # Apply soft thresholding to the new weights with optimal delta
        new_weights = np.clip(new_weights - self.tau[i] / 2, 0, None) - np.clip(
            -new_weights - self.tau[i] / 2, 0, None
        )
        # Return the new weights
        return new_weights



# from typing import Union
#
# import numpy as np
#
# from cca_zoo.linear._iterative._base import BaseIterative
#
#
# class IncrementalPLS(BaseIterative):
#     r"""
#     A class used to fit Incremental PLS
#
#     Parameters
#     ----------
#     latent_dimensions : int, optional
#         Number of latent dimensions to use, by default 1
#     copy_data : bool, optional
#         Whether to copy the data, by default True
#     random_state : int, optional
#         Random state to use, by default None
#     accept_sparse : bool, optional
#         Whether to accept sparse data, by default None
#     batch_size : int, optional
#         Batch size to use, by default 1
#     epochs : int, optional
#         Number of epochs to use, by default 1
#     simple : bool, optional
#         Whether to use the simple update, by default False
#
#     References
#     ----------
#     Arora, Raman, et al. "Stochastic optimization for PCA and PLS." 2012 50th Annual Allerton Conference on Communication, Control, and Computing (Allerton). IEEE, 2012.
#     """
#
#     def __init__(
#         self,
#         latent_dimensions: int = 1,
#         copy_data=True,
#         random_state=None,
#         accept_sparse=None,
#         batch_size=1,
#         epochs=1,
#         simple=False,
#         initialization: Union[str, callable] = "random",
#     ):
#         super().__init__(
#             latent_dimensions=latent_dimensions,
#             copy_data=copy_data,
#             accept_sparse=accept_sparse,
#             random_state=random_state,
#             batch_size=batch_size,
#             epochs=epochs,
#             initialization=initialization,
#         )
#         self.simple = simple
#
#     def _update(self, views):
#         if not hasattr(self, "S"):
#             self.S = np.zeros(self.latent_dimensions)
#             self.count = 0
#         if self.simple:
#             self.simple_update(views)
#         else:
#             self.incremental_update(views)
#         return False
#
#     def incremental_update(self, views):
#         hats = np.stack([view @ weight for view, weight in zip(views, self.weights)])
#         orths = [
#             view - hat @ weight.T
#             for view, weight, hat in zip(views, self.weights, hats)
#         ]
#         self.incrsvd(hats, orths)
#
#     def simple_update(self, views):
#         if not hasattr(self, "M"):
#             self.M = np.zeros((views[0].shape[1], views[1].shape[1]))
#         self.M = (
#             views[0].T @ views[1]
#             + self.weights[0] @ np.diag(self.S) @ self.weights[1].T
#         )
#         U, S, Vt = np.linalg.svd(self.M)
#         self.weights[0] = U[:, : self.latent_dimensions]
#         self.weights[1] = Vt.T[:, : self.latent_dimensions]
#         self.S = S[: self.latent_dimensions]
#
#     def incrsvd(self, hats, orths):
#         Q = np.vstack(
#             (
#                 np.hstack(
#                     (
#                         np.diag(self.S) + hats[0].T @ hats[1],
#                         np.linalg.norm(orths[1], axis=1).T * hats[0].T,
#                     )
#                 ),
#                 np.hstack(
#                     (
#                         (np.linalg.norm(orths[0], axis=1).T * hats[1].T).T,
#                         np.atleast_2d(
#                             np.linalg.norm(orths[0], axis=1, keepdims=True)
#                             @ np.linalg.norm(orths[1], axis=1, keepdims=True).T
#                         ),
#                     )
#                 ),
#             )
#         )
#         U, S, Vt = np.linalg.svd(Q)
#         self.weights[0] = (
#             np.hstack((self.weights[0], orths[0].T / np.linalg.norm(orths[0])))
#             @ U[:, : self.latent_dimensions]
#         )
#         self.weights[1] = (
#             np.hstack((self.weights[1], orths[1].T / np.linalg.norm(orths[1])))
#             @ Vt.T[:, : self.latent_dimensions]
#         )
#         self.S = S[: self.latent_dimensions]



import torch

from cca_zoo.linear._gradient._ey import CCA_EY


class CCA_SVD(CCA_EY):
    def _more_tags(self):
        return {"multiview": False, "stochastic": True}

    def loss(self, views, independent_views=None, **kwargs):
        z = self(views)
        C = torch.cov(torch.hstack(z).T)
        latent_dims = z[0].shape[1]

        Cxy = (C[:latent_dims, latent_dims:] + C[latent_dims:, :latent_dims]) / 2
        Cxx = C[:latent_dims, :latent_dims]

        if independent_views is None:
            Cyy = C[latent_dims:, latent_dims:]
        else:
            independent_z = self(independent_views)
            Cyy = torch.cov(torch.hstack(independent_z).T)[latent_dims:, latent_dims:]

        rewards = torch.trace(2 * Cxy)
        penalties = torch.trace(Cxx @ Cyy)

        return {
            "loss": -rewards + penalties,
            "rewards": rewards,
            "penalties": penalties,
        }


class PLS_SVD(CCA_SVD):
    def loss(self, views, independent_views=None, **kwargs):
        z = self(views)
        C = torch.cov(torch.hstack(z).T)
        latent_dims = z[0].shape[1]

        n = z[0].shape[0]
        Cxy = C[:latent_dims, latent_dims:]
        Cxx = self.torch_weights[0].T @ self.torch_weights[0] / n
        Cyy = self.torch_weights[1].T @ self.torch_weights[1] / n

        rewards = torch.trace(2 * Cxy)
        penalties = torch.trace(Cxx @ Cyy)

        return {
            "loss": -rewards + penalties,
            "rewards": rewards,
            "penalties": penalties,
        }



from typing import Iterable, Union

import numpy as np

from cca_zoo.linear._mcca import MCCA


class PCACCA(MCCA):
    """
    Principal Component Analysis CCA

    Data driven PCA on each view followed by CCA on the PCA components. Keep percentage of variance

    Examples
    --------
    >>> from cca_zoo.linear import PCACCA
    >>> import numpy as np
    >>> X1 = np.random.rand(100, 5)
    >>> X2 = np.random.rand(100, 5)
    >>> model = PCACCA(latent_dimensions=1, percent_variance=0.5)
    >>> model.fit([X1, X2])
    """

    def __init__(
        self,
        latent_dimensions=1,
        copy_data=True,
        random_state=None,
        c: Union[Iterable[float], float] = None,
        eps=1e-9,
        percent_variance=0.99,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            c=c,
            eps=eps,
        )
        # ensure that the percent variance is less than 1 and give a warning if not
        assert percent_variance <= 1, "percent_variance must be less than 1"
        self.percent_variance = percent_variance

    def _process_data(self, views, **kwargs):
        views = self._apply_pca(views)
        for i, view in enumerate(views):
            # Keep the components that explain the percentage of variance
            explained_variance = self.pca_models[i].explained_variance_ratio_
            n_components_ = (
                np.where(np.cumsum(explained_variance) >= self.percent_variance)[0][0]
                + 1
            )
            self.pca_models[i].n_components_ = n_components_
            self.pca_models[i].components_ = self.pca_models[i].components_[
                :n_components_
            ]
            self.pca_models[i].explained_variance_ = self.pca_models[
                i
            ].explained_variance_[:n_components_]
            self.pca_models[i].explained_variance_ratio_ = self.pca_models[
                i
            ].explained_variance_ratio_[:n_components_]
            self.pca_models[i].singular_values_ = self.pca_models[i].singular_values_[
                :n_components_
            ]
        return [
            view[:, : self.pca_models[i].n_components_] for i, view in enumerate(views)
        ]



from cca_zoo.utils.check_values import check_arviz_support


class WeightInferenceDisplay:
    """
    Class for displaying inference-related plots.

    Attributes
    ----------
    idata : arviz.InferenceData
        The posterior samples.
    true_features: array-like, optional
        The true features for comparison in the plot, defaults to None.
    num_views: int, optional
        The number of views, defaults to 2.

    """

    def __init__(self, idata, num_views=2, true_features=None):
        """
        Initialize the WeightInferenceDisplay object.

        Parameters
        ----------
        idata : arviz.InferenceData
            The posterior samples.
        num_views : int, optional
            The number of views, defaults to 2.
        true_features : array-like, optional
            The true features for comparison in the plot, defaults to None.
        """
        self.idata = idata
        self.true_features = true_features
        self.num_views = num_views

    def _validate_plot_params(self):
        """
        Internal method to validate plotting parameters.
        Currently, it checks if arviz is supported.
        """
        check_arviz_support("CorrelationHeatmapDisplay")

    @classmethod
    def from_estimator(cls, pcca_estimator, true_features=None):
        """
        Class method to create an InferenceDisplay instance from an estimator.

        Parameters
        ----------
        pcca_estimator : object
            The estimator object with an 'mcmc' attribute.
        true_features : array-like, optional
            The true features for comparison in the plot, defaults to None.

        Returns
        -------
        WeightInferenceDisplay
            An InferenceDisplay instance.
        """
        return cls.from_mcmc(pcca_estimator.mcmc, true_features)

    @classmethod
    def from_mcmc(cls, mcmc, true_features=None):
        """
        Class method to create an InferenceDisplay instance from mcmc samples.

        Parameters
        ----------
        mcmc : object
            The mcmc samples.
        true_features : array-like, optional
            The true features for comparison in the plot, defaults to None.

        Returns
        -------
        WeightInferenceDisplay
            An InferenceDisplay instance.
        """
        import arviz as az

        idata = az.from_numpyro(mcmc)
        return cls(idata, 2, true_features)

    def plot(self):
        """
        Plot the posterior distributions of parameters and latent variables.
        Adds true values if they are provided.
        """
        import arviz as az
        import matplotlib.pyplot as plt

        for view in range(self.num_views):
            # Plot the posterior distribution of W_0 parameter (for just the first latent variable).
            # Label the weights with their weight index. Make all parameters share x axis.
            trace_plot = az.plot_trace(
                self.idata, var_names=[f"W_{view}"], compact=False, divergences=None
            )

            # For each w in W_0, plot the true value from data.true_features[0]
            for i, ax in enumerate(trace_plot[:, 0]):
                if self.true_features is not None:
                    ax.axvline(
                        self.true_features[view].ravel()[i],
                        color="red",
                        linestyle="--",
                        label="True Value",
                    )
                ax.legend()

            plt.suptitle(f"Posterior Distribution of W_{view}")
            plt.tight_layout()

        plt.show()



from typing import Iterable

import torch

from .._base import BaseDeep
from .._generative._base import _GenerativeMixin


class DVCCA(BaseDeep, _GenerativeMixin):
    """
    A class used to fit a DVCCA model.

    References
    ----------
    Wang, Weiran, et al. 'Deep variational canonical correlation analysis.' arXiv preprint arXiv:1610.03454 (2016).
    https: // arxiv.org / pdf / 1610.03454.pdf
    https: // github.com / pytorch / examples / blob / master / vae / main.py

    """

    def __init__(
        self,
        latent_dimensions: int,
        encoders=None,
        decoders=None,
        private_encoders: Iterable = None,
        latent_dropout=0,
        img_dim=None,
        recon_loss_type="mse",
        **kwargs,
    ):
        super().__init__(latent_dimensions=latent_dimensions, **kwargs)
        self.img_dim = img_dim
        self.latent_dropout = torch.nn.Dropout(p=latent_dropout)
        self.encoders = torch.nn.ModuleList(encoders)
        self.decoders = torch.nn.ModuleList(decoders)
        if private_encoders:
            self.private_encoders = torch.nn.ModuleList(private_encoders)
        else:
            self.private_encoders = None
        self.recon_loss_type = recon_loss_type

    def forward(self, views, mle=True, **kwargs):
        """
        Forward method for the model. Outputs latent encoding for each view

        :param views:
        :param kwargs:
        :return:
        """
        z = {}
        # Used when we get reconstructions
        z["mu_shared"], z["logvar_shared"] = self._encode(views)
        z["shared"] = self._sample(z["mu_shared"], z["logvar_shared"], mle)
        if self.private_encoders is not None:
            z["mu_private"], z["logvar_private"] = self._encode_private(views)
            z["private"] = [
                self._sample(mu_, logvar_, mle)
                for mu_, logvar_ in zip(z["mu_private"], z["logvar_private"])
            ]
        return z

    def _sample(self, mu, logvar, mle):
        """

        :param mu:
        :param logvar:
        :param mle: whether to return the maximum likelihood (i.e. mean) or whether to sample
        :return: a sample from latent vector
        """
        if mle:
            return mu
        else:
            return mu + torch.randn_like(logvar) * torch.exp(0.5 * logvar)

    def _encode(self, views):
        mu = []
        logvar = []
        for i, encoder in enumerate(self.encoders):
            mu_i, logvar_i = encoder(views[i])
            mu.append(mu_i)
            logvar.append(logvar_i)
        return torch.stack(mu).sum(dim=0), torch.stack(logvar).sum(dim=0)

    def _encode_private(self, views):
        mu = []
        logvar = []
        for i, private_encoder in enumerate(self.private_encoders):
            mu_i, logvar_i = private_encoder(views[i])
            mu.append(mu_i)
            logvar.append(logvar_i)
        return mu, logvar

    def _decode(self, z, uncertainty=False, **kwargs):
        x = []
        if uncertainty:
            z["shared"] = z["logvar_shared"]
            if self.private_encoders is not None:
                z["private"] = z["logvar_private"]
        for i, decoder in enumerate(self.decoders):
            if "private" in z:
                x_i = decoder(
                    torch.cat(
                        (
                            self.latent_dropout(z["shared"]),
                            self.latent_dropout(z["private"][i]),
                        ),
                        dim=-1,
                    )
                )
            else:
                x_i = decoder(self.latent_dropout(z["shared"]))
            x.append(x_i)
        return x

    def loss(self, batch, **kwargs):
        z = self(batch["views"], mle=False)
        recons = self._decode(z)
        loss = dict()
        loss["reconstruction"] = torch.stack(
            [
                self.recon_loss(x, recon, loss_type=self.recon_loss_type)
                for x, recon in zip(batch["views"], recons)
            ]
        ).sum()
        loss["kl shared"] = (
            self.kl_loss(z["mu_shared"], z["logvar_shared"]) / batch["views"][0].numel()
        )
        if "private" in z:
            loss["kl private"] = torch.stack(
                [
                    self.kl_loss(mu_, logvar_) / batch["views"][0].numel()
                    for mu_, logvar_ in zip(z["mu_private"], z["logvar_private"])
                ]
            ).sum()
        loss["objective"] = torch.stack(tuple(loss.values())).sum()
        return loss

    def transform(
        self,
        loader: torch.utils.data.DataLoader,
    ):
        """
        :param loader: a dataloader that matches the structure of that used for training
        :return: transformed views
        """
        with torch.no_grad():
            z_shared = []
            z_private = []
            for batch_idx, batch in enumerate(loader):
                views = [view.to(self.device) for view in batch["views"]]
                z_ = self(views)
                z_shared.append(z_["shared"].cpu())
                if "private" in z_:
                    z_private.append(self.detach_all(z_["private"]))
        z = {"shared": torch.vstack(z_shared).cpu().numpy()}
        if "private" in z_:
            z["private"] = [torch.vstack(i).cpu().numpy() for i in zip(*z_private)]
        return z
